@@ -80,7 +80,7 @@ static void as86_write_section (struct Section *, int);
 static int as86_add_string (char *name);
 static void as86_sect_write(struct Section *, unsigned char *, unsigned long);
 
-static void as86_init(FILE *fp, efunc errfunc, ldfunc ldef) {
+static void as86_init(FILE *fp, efunc errfunc, ldfunc ldef, evalfunc eval) {
     as86fp = fp;
     error = errfunc;
     (void) ldef;		       /* placate optimisers */
@@ -158,10 +158,15 @@ static int as86_add_string (char *name) {
 }
 
 static void as86_deflabel (char *name, long segment, long offset,
-			   int is_global) {
+			   int is_global, char *special) {
     struct Symbol *sym;
 
-    if (name[0] == '.' && name[1] == '.') {
+    if (special)
+	error (ERR_NONFATAL, "as86 format does not support any"
+	       " special symbol types");
+
+    if (name[0] == '.' && name[1] == '.' && name[2] != '@') {
+	error (ERR_NONFATAL, "unrecognised special symbol `%s'", name);
 	return;
     }
 
@@ -354,7 +359,7 @@ static void as86_write(void) {
 	    sym->flags |= 0 << 14, symlen += 4;
 	else if (sym->value >= 0 && sym->value <= 255)
 	    sym->flags |= 1 << 14, symlen += 5;
-	else if (sym->value >= 0 && sym->value <= 65535)
+	else if (sym->value >= 0 && sym->value <= 65535L)
 	    sym->flags |= 2 << 14, symlen += 6;
 	else
 	    sym->flags |= 3 << 14, symlen += 8;
@@ -365,31 +370,31 @@ static void as86_write(void) {
      * descriptor word at the same time.
      */
     seglen = segsize = 0;
-    if ((unsigned long) stext.len > 65535)
-	segsize |= 0x03000000, seglen += 4;
+    if ((unsigned long) stext.len > 65535L)
+	segsize |= 0x03000000L, seglen += 4;
     else
-	segsize |= 0x02000000, seglen += 2;
-    if ((unsigned long) sdata.len > 65535)
-	segsize |= 0xC0000000, seglen += 4;
+	segsize |= 0x02000000L, seglen += 2;
+    if ((unsigned long) sdata.len > 65535L)
+	segsize |= 0xC0000000L, seglen += 4;
     else
-	segsize |= 0x80000000, seglen += 2;
+	segsize |= 0x80000000L, seglen += 2;
 
     /*
      * Emit the as86 header.
      */
-    fwritelong (0x000186A3, as86fp);
+    fwritelong (0x000186A3L, as86fp);
     fputc (0x2A, as86fp);
     fwritelong (27+symlen+seglen+strslen, as86fp);   /* header length */
     fwritelong (stext.len+sdata.len, as86fp);
     fwriteshort (strslen, as86fp);
     fwriteshort (0, as86fp);	       /* class = revision = 0 */
-    fwritelong (0x55555555, as86fp);   /* segment max sizes: always this */
+    fwritelong (0x55555555L, as86fp);   /* segment max sizes: always this */
     fwritelong (segsize, as86fp);      /* segment size descriptors */
-    if (segsize & 0x01000000)
+    if (segsize & 0x01000000L)
 	fwritelong (stext.len, as86fp);
     else
 	fwriteshort (stext.len, as86fp);
-    if (segsize & 0x40000000)
+    if (segsize & 0x40000000L)
 	fwritelong (sdata.len, as86fp);
     else
 	fwriteshort (sdata.len, as86fp);
@@ -428,7 +433,7 @@ static void as86_write(void) {
 static void as86_set_rsize (int size) {
     if (as86_reloc_size != size) {
 	switch (as86_reloc_size = size) {
-	  case 1: fputc (0x01, as86fp); break;   /* shouldn't happen */
+	  case 1: fputc (0x01, as86fp); break;
 	  case 2: fputc (0x02, as86fp); break;
 	  case 4: fputc (0x03, as86fp); break;
 	  default: error (ERR_PANIC, "bizarre relocation size %d", size);
@@ -481,7 +486,7 @@ static void as86_write_section (struct Section *sect, int index) {
 	     */
 	    as86_set_rsize (p->bytes);
 	    s = p->offset;
-	    if (s > 65535)
+	    if (s > 65535L)
 		s = 3;
 	    else if (s > 255)
 		s = 2;
@@ -532,9 +537,15 @@ static void as86_filename (char *inname, char *outname, efunc error) {
     standard_extension (inname, outname, ".o", error);
 }
 
+static char *as86_stdmac[] = {
+    "%define __SECT__ [section .text]",
+    NULL
+};
+
 struct ofmt of_as86 = {
     "Linux as86 (bin86 version 0.3) object files",
     "as86",
+    as86_stdmac,
     as86_init,
     as86_out,
     as86_deflabel,
