@@ -11,21 +11,26 @@
  * May require the system to make an extra pass of the modules to be
  * loaded eliminating those that aren't required.
  *
- * Support libaries (.a files - requires a 'ranlib' type utility)
+ * Support all the existing documented options...
  *
- * -s option to strip resolved symbols from exports.
+ * Support libaries (.a files - requires a 'ranlib' type utility)
+ *		(I think I've got this working, so I've upped the version)
+ *
+ * -s option to strip resolved symbols from exports. (Could make this an
+ * external utility)
  */
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 
-#include "nasm.h"
 #include "rdoff.h"
 #include "nasmlib.h"
 #include "symtab.h"
 #include "collectn.h"
+#include "rdlib.h"
 
-#define LDRDF_VERSION "0.11"
+#define LDRDF_VERSION "0.30"
 
 /* global variables - those to set options: */
 
@@ -46,8 +51,17 @@ struct modulenode {
     struct modulenode *next;
 };
 
+#define newstr(str) strcpy(malloc(strlen(str) + 1),str)
+#define newstrcat(s1,s2) strcat(strcpy(malloc(strlen(s1)+strlen(s2)+1),s1),s2)
+
+
 struct modulenode *modules = NULL,*lastmodule = NULL;
 
+/* the linked list of libraries to be searched for missing imported
+   symbols */
+
+struct librarynode * libraries = NULL, * lastlib = NULL;
+
 void *symtab;	/* The symbol table */
 
 rdf_headerbuf * newheader ;	/* New header to be written to output */
@@ -101,6 +115,218 @@ void loadmodule(char *filename)
 	   lastmodule->coderel,lastmodule->f.code_len,
 	   lastmodule->datarel,lastmodule->f.data_len);
 
+  lastmodule->header = malloc(lastmodule->f.header_len);
+  if (!lastmodule->header) {
+      fprintf(stderr,"ldrdf: out of memory\n");
+      exit(1);
+  }
+
+  if (rdfloadseg(&lastmodule->f,RDOFF_HEADER,lastmodule->header))
+  {
+      rdfperror("ldrdf",filename);
+      exit(1);
+  }
+}
+
+/* load_library		add a library to list of libraries to search
+ *                      for undefined symbols
+ */
+
+void load_library(char * name)
+{
+    if (verbose)
+	printf("adding library %s to search path\n",name);
+	
+    if (! lastlib) {
+	lastlib = libraries = malloc(sizeof(struct librarynode));
+    }
+    else
+    {
+	lastlib->next = malloc(sizeof(struct librarynode));
+	lastlib = lastlib->next;
+    }
+
+    if (! lastlib) {
+	fprintf(stderr, "ldrdf: out of memory\n");
+	exit(1);
+    }
+    strcpy (lastlib->name = malloc (1+strlen(name)), name);
+    lastlib->fp = NULL;
+    lastlib->referenced = 0;
+    lastlib->next = NULL;
+}
+
+
+/* build_symbols()	step through each module's header, and locate
+ *			exported symbols, placing them in a global table
+ */
+
+long bsslength;
+
+void mod_addsymbols(struct modulenode * mod)
+{
+    rdfheaderrec *r;
+    symtabEnt e;
+    long cbBss;
+
+    mod->bssrel = bsslength;
+    cbBss = 0;
+    rdfheaderrewind(&mod->f);
+    while ((r = rdfgetheaderrec(&mod->f)))
+    {
+
+	if (r->type == 5)		/* Allocate BSS */
+	    cbBss += r->b.amount;
+
+	if (r->type != 3) continue;	/* ignore all but export recs */
+
+	e.segment = r->e.segment;
+	e.offset = r->e.offset + 
+	    (e.segment == 0 ? mod->coderel : /* 0 -> code */
+	     e.segment == 1 ? mod->datarel : /* 1 -> data */
+	     mod->bssrel) ; /* 2 -> bss  */
+	
+	e.flags = 0;
+	e.name = malloc(strlen(r->e.label) + 1);
+	if (! e.name)
+	{
+	    fprintf(stderr,"ldrdf: out of memory\n");
+	    exit(1);
+	}
+	strcpy(e.name,r->e.label);
+	symtabInsert(symtab,&e);
+    }
+    bsslength += cbBss;
+}
+
+void build_symbols()
+{
+  struct modulenode *mod;
+
+  if (verbose) printf("building global symbol table:\n");
+  newheader = rdfnewheader();
+
+  symtab = symtabNew();
+  bsslength = 0;		  /* keep track of location of BSS symbols */
+
+  for (mod = modules; mod; mod = mod->next)
+  {
+      mod_addsymbols( mod );
+  }
+  if (verbose)
+  {
+      symtabDump(symtab,stdout);
+      printf("BSS length = %ld bytes\n\n",bsslength);
+  }
+}
+
+
+/* scan_libraries()	search through headers of modules for undefined
+ *			symbols, and scan libraries for those symbols,
+ *			adding library modules found to list of modules
+ *			to load.					*/
+
+void scan_libraries(void)
+{
+    struct modulenode 	* mod, * nm;
+    struct librarynode	* lib;
+    rdfheaderrec	* r;
+    int			found;
+    char		* tmp;
+
+    if (verbose) printf("Scanning libraries for unresolved symbols...\n");
+
+    mod = modules;
+
+    while (mod)
+    {
+	rdfheaderrewind(&mod->f);
+
+	while ((r = rdfgetheaderrec(&mod->f)))
+	{
+	    if (r->type != 2) continue;	/* not an import record */
+	    if ( symtabFind (symtab,r->i.label) )
+		continue;		/* symbol already defined */
+	    
+	    /* okay, we have an undefined symbol... step through
+	       the libraries now */
+	    if (verbose >= 2) {
+		printf("undefined symbol '%s'...",r->i.label);
+		fflush(stdout);
+	    }
+
+	    lib = libraries;
+	    found = 0;
+
+	    tmp = newstr(r->i.label);
+	    while (! found && lib)
+	    {
+		/* move this to an outer loop...! */
+		nm = malloc(sizeof(struct modulenode));
+
+		if (rdl_searchlib(lib,tmp,&nm->f))
+		{	/* found a module in the library */
+
+		    /* create a modulenode for it */
+
+		    if (! nm) {
+			fprintf(stderr,"ldrdf: out of memory\n");
+			exit(1);
+		    }
+
+		    nm->name = newstrcat(lib->name,nm->f.name);
+		    if (verbose >= 2) printf("found in '%s'\n",nm->name);
+
+		    nm->coderel = lastmodule->coderel + lastmodule->f.code_len;
+		    if (nm->coderel % align != 0)
+			nm->coderel += align - (nm->coderel % align);
+
+		    nm->datarel = lastmodule->datarel + lastmodule->f.data_len;
+		    if (nm->datarel % align != 0)
+			nm->datarel += align - (nm->datarel % align);
+
+		    nm->header = malloc(nm->f.header_len);
+		    if (! nm->header)
+		    {
+			fprintf(stderr,"ldrdf: out of memory\n");
+			exit(1);
+		    }
+
+		    if (rdfloadseg(&nm->f,RDOFF_HEADER,nm->header))
+		    {
+			rdfperror("ldrdf",nm->name);
+			exit(1);
+		    }
+
+		    nm->next = NULL;
+		    found = 1;
+		    lastmodule->next = nm;
+		    lastmodule = nm;
+
+		    if (verbose)
+			printf("%s code = %08lx (+%04lx), data = %08lx "
+			       "(+%04lx)\n",lastmodule->name,
+			       lastmodule->coderel,lastmodule->f.code_len,
+			       lastmodule->datarel,lastmodule->f.data_len);
+
+		    /* add the module's info to the symbol table */
+		    mod_addsymbols(nm);
+		}
+		else
+		{
+		    if (rdl_error) {
+			rdl_perror("ldrdf",lib->name);
+			exit(1);
+		    }
+		    free(nm);
+		}
+		lib = lib->next;
+	    }
+	    free(tmp);
+	    if (!found && verbose >= 2) printf("not found\n");
+	}
+	mod = mod->next;
+    }
 }
 
 /* load_segments()	allocates memory for & loads the code & data segs
@@ -108,7 +334,7 @@ void loadmodule(char *filename)
  */
 
 char *text,*data;
-long textlength,datalength,bsslength;
+long textlength,datalength;
 
 void load_segments(void)
 {
@@ -150,77 +376,17 @@ void load_segments(void)
 
   mod = modules;
   while (mod) {		/* load the segments for each module */
-    mod->header = malloc(mod->f.header_len);
-    if (!mod->header) {
-      fprintf(stderr,"ldrdf: out of memory\n");
-      exit(1);
-    }
-    if (rdfloadseg(&mod->f,RDOFF_HEADER,mod->header) ||
-	rdfloadseg(&mod->f,RDOFF_CODE,&text[mod->coderel]) ||
-	rdfloadseg(&mod->f,RDOFF_DATA,&data[mod->datarel])) {
-      rdfperror("ldrdf",mod->name);
-      exit(1);
-    }
-    rdfclose(&mod->f);	/* close file; segments remain */
-    mod = mod->next;
-  }
-}
-
-/* build_symbols()	step through each module's header, and locate
- *			exported symbols, placing them in a global table
- */
-
-void build_symbols()
-{
-  struct modulenode *mod;
-  rdfheaderrec *r;
-  symtabEnt e;
-  long bssloc,cbBss;
-
-  if (verbose) printf("building global symbol table:\n");
-  newheader = rdfnewheader();
-
-  symtab = symtabNew();
-  bssloc = 0;			/* keep track of location of BSS symbols */
-
-  for (mod = modules; mod; mod = mod->next)
-  {
-      mod->bssrel = bssloc;
-      cbBss = 0;
-      rdfheaderrewind(&mod->f);
-      while ((r = rdfgetheaderrec(&mod->f)))
-      {
-
-	  if (r->type == 5)		/* Allocate BSS */
-	      cbBss += r->b.amount;
-
-	  if (r->type != 3) continue;	/* ignore all but export recs */
-
-	  e.segment = r->e.segment;
-	  e.offset = r->e.offset +
-	               (e.segment == 0 ? mod->coderel : /* 0 -> code */
-			e.segment == 1 ? mod->datarel : /* 1 -> data */
-			                 mod->bssrel) ; /* 2 -> bss  */
-	  e.flags = 0;
-	  e.name = malloc(strlen(r->e.label) + 1);
-	  if (! e.name)
-	  {
-	      fprintf(stderr,"ldrdf: out of memory\n");
-	      exit(1);
-	  }
-	  strcpy(e.name,r->e.label);
-	  symtabInsert(symtab,&e);
+      if (verbose >= 2) printf("  loading %s\n",mod->name);
+      if (rdfloadseg(&mod->f,RDOFF_CODE,&text[mod->coderel]) ||
+	  rdfloadseg(&mod->f,RDOFF_DATA,&data[mod->datarel])) {
+	  rdfperror("ldrdf",mod->name);
+	  exit(1);
       }
-      bssloc += cbBss;
+      rdfclose(&mod->f);	/* close file; segments remain */
+      mod = mod->next;
   }
-  if (verbose)
-  {
-      symtabDump(symtab,stdout);
-      printf("BSS length = %ld bytes\n\n",bssloc);
-  }
-  bsslength = bssloc;
 }
-
+    
 /* link_segments()	step through relocation records in each module's
  *			header, fixing up references.
  */
@@ -230,7 +396,7 @@ void link_segments(void)
   struct modulenode	*mod;
   Collection		imports;
   symtabEnt		*s;
-  long 			rel,relto = 0; /* placate gcc */
+  long 			rel,relto;
   char 			*seg;
   rdfheaderrec		*r;
   int			bRelative;
@@ -243,6 +409,7 @@ void link_segments(void)
     if (verbose >= 2) printf("* processing %s\n",mod->name);
     rdfheaderrewind(&mod->f);
     while((r = rdfgetheaderrec(&mod->f))) {
+	if (verbose >= 3) printf("record type: %d\n",r->type);
 	switch(r->type) {
 	case 1:		/* relocation record */
 	    if (r->r.segment >= 64) {     	/* Relative relocation; */
@@ -251,8 +418,11 @@ void link_segments(void)
 		relto = r->r.segment == 0 ? mod->coderel : mod->datarel;
 	    }
 	    else
+	    {
 		bRelative = 0;		/* non-relative - need to relocate
 					 * at load time			*/
+		relto = 0;	       /* placate optimiser warnings */
+	    }
 
 	    /* calculate absolute offset of reference, not rel to beginning of
 	       segment */
@@ -275,19 +445,19 @@ void link_segments(void)
 		    break;
 		}
 		rel = s->offset;
-
-		r->r.refseg = s->segment;	/* change referred segment,
+	  
+		r->r.refseg = s->segment;	/* change referred segment, 
 						   so that new header is
 						   correct */
 	    }
 
 	    if (bRelative)	/* Relative - subtract current segment start */
-		rel -= relto;
-	    else
+		rel -= relto; 
+	    else  
 	    {			/* Add new relocation header */
 		rdfaddheader(newheader,r);
 	    }
-
+	    
 	    /* Work out which segment we're making changes to ... */
 	    if (r->r.segment == 0) seg = text;
 	    else if (r->r.segment == 1) seg = data;
@@ -326,7 +496,7 @@ void link_segments(void)
 			r->i.label,mod->name);
 		errors = 1;
 	    }
-	    else
+	    else 
 	    {
 		*colln(&imports,r->i.segment - 2) = s;
 		if (verbose >= 2)
@@ -336,8 +506,11 @@ void link_segments(void)
 
 	case 3:		/* export; dump to output new version */
 	    s = symtabFind(symtab, r->e.label);
-	    if (! s) continue;	/* eh? probably doesn't matter... */
-
+	    if (! s) {
+		fprintf(stderr,"ldrdf: internal error - undefined symbol %s "
+			"exported in header of '%s'\n",r->e.label,mod->name);
+		continue;
+	    }
 	    r->e.offset = s->offset;
 	    rdfaddheader(newheader,r);
 	    break;
@@ -347,10 +520,14 @@ void link_segments(void)
 	    break;
 	}
     }
+    if (rdf_errno != 0) {
+	rdfperror("ldrdf",mod->name);
+	exit(1);
+    }
     collection_reset(&imports);
   }
 }
-
+    
 /* write_output()	write linked program out to a file */
 
 void write_output(char *filename)
@@ -358,14 +535,16 @@ void write_output(char *filename)
     FILE		* fp;
     rdfheaderrec	r;
 
+    if (verbose) printf("writing output to '%s'\n",filename);
+
     fp = fopen(filename,"wb");
     if (! fp)
     {
 	fprintf(stderr,"ldrdf: could not open '%s' for writing\n",filename);
 	exit(1);
     }
-
-
+  
+    
     /* add BSS length count to header... */
     if (bsslength)
     {
@@ -412,7 +591,8 @@ void write_output(char *filename)
  */
 
 const char *usagemsg = "usage:\n"
-" ldrdf [-o outfile | -x] [-a x] [-v] [-p x] [--] infile [infile ...]\n\n"
+" ldrdf [-o outfile | -x] [-a x] [-v] [-p x] [--] infile [infile ...]\n"
+"       [-l<libname> ...]\n\n"
 " ldrdf -h	displays this message\n"
 " ldrdf -r	displays version information\n\n"
 "   -o selects output filename (default is aout.rdx)\n"
@@ -420,7 +600,9 @@ const char *usagemsg = "usage:\n"
 "   -a x causes object program to be statically relocated to address 'x'\n"
 "   -v turns on verbose mode\n"
 "   -p x causes segments to be aligned (padded) to x byte boundaries\n"
-"      (default is 16 bytes)\n";
+"      (default is 16 bytes)\n"
+"   -l<name> causes 'name' to be linked in as a library. Note no search is\n"
+"      performed - the entire pathname MUST be specified.\n";
 
 void usage(void)
 {
@@ -499,6 +681,9 @@ int main(int argc,char **argv)
       }
       if (verbose > 1) printf("alignment %d selected\n",align);
     }
+    else if (procsw && !strncmp(*argv,"-l",2)) {
+	load_library(*argv + 2);
+    }
     else if (procsw && !strcmp(*argv,"--")) {
       procsw = 0;
     }
@@ -512,12 +697,15 @@ int main(int argc,char **argv)
      unreferenced modules from the list of modules here, so that
      we know about the final size once libraries have been linked in */
 
+  build_symbols();	/* build a global symbol table...           */
+
+  scan_libraries();	/* check for imported symbols not in table,
+			   and ensure the relevant library modules
+			   are loaded */
+
   load_segments();	/* having calculated size of reqd segments, load
 			   each rdoff module's segments into memory */
 
-  build_symbols();	/* build a global symbol table...
-			   perhaps this should be done before load_segs? */
-
   link_segments();	/* step through each module's header, and resolve
 			   references to the global symbol table.
 			   This also does local address fixups. */
